// Generation.h: interface for the Generation class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_)
#define AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include	"Random.h"

// Warning!  These numbers are NOT chosen arbitrarily. They must be
// related by POP_SIZE choose 2 = PARENT_SIZE.  That is, the formula:
//	               PARENT_SIZE!
//	POP_SIZE = -----------------------
//	           2! * (PARENT_SIZE - 2)!
#define	POP_SIZE	105
#define	PARENT_SIZE	15
#define	ROWS		8
#define COLS		8

// The Generation object holds a big array of T objects and provides
// methods for creating and advancing a population

template <class T>
class Generation
{
public:
	Generation();
	~Generation();

	void ChooseParents(void);
	int operator++(int);
	void SetMutationRate(int rate);
	void Log(void);
	T GetBest(void);
	void Init(void);

private:
	int		genCnt;

	int		Fitness[POP_SIZE];			// All member fitnesses
	int		mutationRate;				// Chars changed per generation

	int		Parent[PARENT_SIZE];		// Array of indexes into the population
	T		evenPopulation[POP_SIZE];	// The members of the population
	T		oddPopulation[POP_SIZE];

	T		bestEver;					// The best member EVER seen
	int		bestEverFitness;			// The fitness of best EVER

	int		bestOfBreed;				// The fittest in THIS generation
	int		letterCount;				// The letter count of the fittest member

	Random<char>	randomLetter;
	Random<int>		randomMember;
	Random<int>		randomRow;
	Random<int>		randomCol;
};

template <class T>
Generation<T>::Generation()
		: genCnt(0), bestOfBreed(0), bestEverFitness(0),
		  randomLetter('A','Z'), randomMember(0, POP_SIZE-1),
		  randomRow(1, ROWS), randomCol(1, COLS)
{
	// The initialization of both even and odd generations isn't necessary
	// since one will be overwritten.  But since this only occurs once
	// in the program, better safe than sorry.
	for (int i=0; i<POP_SIZE; i++)
	{
		Fitness[i] = 0;
		oddPopulation[i].Init();
		evenPopulation[i].Init();
	}
}

template <class T>
Generation<T>::~Generation()	{}


// There are going to be millions of generations--so performance counts.
// I do every reasonable thing to make each generation faster.
// Here's the general scheme:
// Look through the generation for the fittest members.  Don't copy
// them into an array of parent members (too slow), instead, just
// remember where the good ones are (save the index into parent array).
// Then use these parents to create new members into the other (odd
// or even) generation.  This double generation scheme means the parents
// don't have to be saved from the destruction of overwriting.  Since
// the members have to be written anyway, there is only a space penalty
// but no performance penalty.

// Throughout the generation code there is heavy repetition.
// Almost everything is done for both the even and odd generations.
// The conditionals (even or odd) could have been checked inside the
// loops to make the code much smaller, but performance matters here.

template <class T>
void Generation<T>::ChooseParents(void)
{
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<POP_SIZE; i++)
			Fitness[i] = evenPopulation[i].DetermineFitness();
	}
	else
	{
		for (int i=0; i<POP_SIZE; i++)
			Fitness[i] = oddPopulation[i].DetermineFitness();
	}

	// To find the fittest member in the population, look at
	// the fitness of the first grid, make that a reference,
	// then compare every other grid to it.  When another is
	// fitter, use that as the reference.
	// Do the first separately because special stuff is done to
	// the first one (letters counted and reported).  To prevent
	// the repetition, a conditional could have been stuck in the
	// loop but speed is of the essence.
	bestOfBreed = Fitness[0];
	int index = 0;
	for (int i=1; i<POP_SIZE; i++)
	{
		if (Fitness[i] > bestOfBreed)
		{
			bestOfBreed = Fitness[i];
			index = i;
		}
	}
	// If this is a new record, remember it:
	if (bestOfBreed > bestEverFitness)
	{
		bestEverFitness = bestOfBreed;
		if ((genCnt % 2) == 0)
			bestEver = evenPopulation[index];
		else
			bestEver = oddPopulation[index];
	}
	Parent[0] = index;					// Remember where this parent was
	Fitness[index] = -1;				// Don't pick it again

	if ((genCnt % 2) == 0)
		letterCount = evenPopulation[index].GetLetterCount();
	else
		letterCount = oddPopulation[index].GetLetterCount();

	// Now find the next fittest members.  The most fit had to have
	// its letters counted, but these can just be used as parents.
	int nextBiggest;
	for (int p=1; p<PARENT_SIZE; p++)
	{
		// Find the biggest in the fitness array:
		nextBiggest = Fitness[0];
		index = 0;
		for (int i=1; i<POP_SIZE; i++)
		{
			if (Fitness[i] > nextBiggest)
			{
				nextBiggest = Fitness[i];
				index = i;
			}
		}
		Parent[p] = index;				// Remember where this parent was
		Fitness[index] = -1;			// Don't pick it again
	}
}

// Postfix incrementer--make the next generation of course:
template <class T>
int Generation<T>::operator++ (int)
{
	genCnt++;
	ChooseParents();

	int childCnt = 0;
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<PARENT_SIZE; i++)
		{
			for (int j=i+1; j<PARENT_SIZE; j++)
			{
				// Be fruitful and multiply:
				oddPopulation[childCnt++] = 
					evenPopulation[Parent[i]] * evenPopulation[Parent[j]];
			}
		}
	}
	else
	{
		for (int i=0; i<PARENT_SIZE; i++)
		{
			for (int j=i+1; j<PARENT_SIZE; j++)
			{
				// Be fruitful and multiply:
				evenPopulation[childCnt++] = 
					oddPopulation[Parent[i]] * oddPopulation[Parent[j]];
			}
		}
	}
	// Oh no! Deadly radioactive fallout.  Aaagh, mutants!
	if ((genCnt % 2) == 0)
	{
		for (int i=0; i<mutationRate; i++)
		{
			char c;
			c = randomLetter.Pick();
			oddPopulation[randomMember.Pick()].
				letterGrid[randomRow.Pick()][randomCol.Pick()] = c;
//				randomLetter.Pick();
		}
	}
	else
	{
		for (int i=0; i<mutationRate; i++)
		{
			char c;
			c = randomLetter.Pick();
			evenPopulation[randomMember.Pick()].
				letterGrid[randomRow.Pick()][randomCol.Pick()] = c;
//				randomLetter.Pick();
		}
	}
	return bestOfBreed;
}

template <class T>
void Generation<T>::SetMutationRate(int rate)
{
	mutationRate = rate;
}

// Print:
template <class T>
void Generation<T>::Log(void)
{
//	cout << "  Gen: " << genCnt;
//	cout << "\tWords: " << bestOfBreed;
//	cout << "\tLetters: " << letterCount << "\n";
}

// Get the Best:
template <class T>
T Generation<T>::GetBest(void)
{
	return bestEver;
}

template <class T>
void Generation<T>::Init(void)
{
	bestEverFitness = 0;
	bestEver.Init();
	for (int i=0; i<POP_SIZE; i++)
	{
		Fitness[i] = 0;
		oddPopulation[i].Init();
		evenPopulation[i].Init();
	}
}

#endif // !defined(AFX_GENERATION_H__8FA8C9D4_8B21_11D3_9887_0040056D374E__INCLUDED_)
